from ximilar.client import RestClient
from ximilar.client.constants import RECORDS

SIMILARITY_PHOTOS = 'similarity/photos/v2/'
SIMILARITY_PHOTOS = 'similarity/products/v2/'

SMART_INSERT_ENDPOINT = 'smart/v2/product/insert'
SMART_SEARCH_ENDPOINT = 'smart/v2/product/search'
SMART_DETECT_ENDPOINT = 'smart/v2/product/detect'
SMART_SEARCH_OBJ_ENDPOINT = 'smart/v2/product/search_by_object'


class Similarity(RestClient):
    def insert(self):
        raise NotImplementedError

    def search(self, records, collection):
        raise NotImplementedError

    def delete(self, records, collection):
        raise NotImplementedError

    def random(self, records, collection):
        raise NotImplementedError

    def get(self):
        raise NotImplementedError


class SmartSearchClient(RestClient):
    def __init__(self, token, collection):
        super(SmartSearchClient, self).__init__(token)

        self.max_size = 1000
        self.headers['collection'] = collection

    def insert(self):
        raise NotImplementedError

    def search(self, records, collection):
        raise NotImplementedError

    def detect(self, records):
        """
        Detects Objects and Tags without searching items.
        :param records: list of dictionaries with _url|_file|_base64
        :return: json response
        """
        records = self.preprocess_records(records)
        return self.post(SMART_DETECT_ENDPOINT, data={RECORDS: records})

    def search_by_object(self, records, collection):
        raise NotImplementedError



import requests
import json
import base64
import cv2

from ximilar.client.constants import FILE, BASE64, IMG_DATA


class RestClient(object):
    """
    Parent class that implements HTTP GET, POST, DELETE methods with requests lib and loading images to base64.

    All objects contains TOKEN and ENDPOINT information.
    """
    def __init__(self, token, endpoint='https://api.ximilar.com/'):
        self.token = token
        self.cache = {}
        self.endpoint = endpoint
        self.max_size = 600
        self.headers = {'Content-Type': 'application/json',
                        'Authorization': 'Token ' + self.token}

    def invalidate(self):
        self.cache = {}

    def get(self, api_endpoint, data=None):
        """
        Call the http GET request with data.
        :param api_endpoint: endpoint path
        :param data: optional data
        :return: json response
        """
        result = requests.get(self.endpoint+api_endpoint, headers=self.headers, data=data)
        return result.json()

    def post(self, api_endpoint, data=None, files=None):
        """
        Call the http POST request with data.
        :param api_endpoint: endpoint path
        :param data: optional data
        :param files: optional files to upload
        :return: json response
        """
        self.invalidate()
        if data:
            data = json.dumps(data)

        headers = self.headers if not files else {'Authorization': 'Token ' + self.token}
        result = requests.post(self.endpoint+api_endpoint, headers=headers, data=data, files=files)
        return result.json()

    def delete(self, api_endpoint, data=None):
        """
        Call the http DELETE request with data.
        :param api_endpoint: endpoint path
        :param data: optional data
        :return: response
        """
        self.invalidate()
        return requests.delete(self.endpoint+api_endpoint, headers=self.headers, data=data)

    def resize_image_data(self, image_data, aspect_ratio=False):
        """
        Resize image data that are no bigger than max_size.
        :param image_data: cv2/np ndarray
        :return: cv2/np ndarray
        """
        # do not resize image if set to 0
        if self.max_size == 0:
            return image_data

        height, width, _ = image_data.shape
        if height > self.max_size and width > self.max_size and not aspect_ratio:
            image_data = cv2.resize(image_data, (self.max_size, self.max_size))
        if height > self.max_size and width > self.max_size and aspect_ratio:
            image_data = cv2.resize(image_data, self.get_aspect_ratio_dim(image_data, self.max_size))
        return image_data

    def get_aspect_ratio_dim(self, image, img_size):
        if image.shape[0] > image.shape[1]:
            r = float(img_size) / image.shape[1]
            dim = (img_size, int(image.shape[0] * r))
        else:
            r = float(img_size) / image.shape[0]
            dim = (int(image.shape[1] * r), img_size)
        return dim

    def load_base64_file(self, path):
        """
        Load file from disk to base64.
        :param path: local path to the image
        :return: base64 encoded string
        """
        image = cv2.imread(str(path))
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        image = self.resize_image_data(image)
        image = self.cv2img_to_base64(image)
        return image

    def cv2img_to_base64(self, image):
        """
        Load raw numpy/cv2 data of image to base64. The input image to this method should have RGB order.
        The ximilar accepts base64 data to have BGR order that is why we convert it here.
        The image_data was loaded in similar way:
            image = cv2.imread(str(path))
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        :param image_data: numpy/cv2 data with RGB order
        :return: base64 encoded string
        """
        image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
        image = self.resize_image_data(image)
        retval, buffer = cv2.imencode('.jpg', image)
        jpg_as_text = base64.b64encode(buffer).decode('utf-8')
        return jpg_as_text

    def preprocess_records(self, records):
        """
        Preprocess all records (list of dictionaries with possible '_base64'|'_file'|'_url' fields
        before processing/upload to Ximilar Application.
        :param records: list of dictionaries
        :return: modified list of dictionaries
        """
        for i in range(len(records)):
            if FILE in records[i] and BASE64 not in records[i] and IMG_DATA not in records[i]:
                records[i][BASE64] = self.load_base64_file(records[i][FILE])
            elif IMG_DATA in records[i]:
                records[i][BASE64] = self.cv2img_to_base64(records[i][IMG_DATA])

            # finally we need to delete the image data and just send url or base64
            if IMG_DATA in records[i]:
                del records[i][IMG_DATA]

            if FILE in records[i]:
                del records[i][FILE]

        return records



from .client import RestClient
from .recognition import RecognitionClient
from .tagging import GenericTaggingClient, FashionTaggingClient
from .colors import DominantColorProductClient, DominantColorGenericClient
from .search import SmartSearchClient



import pytest

from ximilar.client.recognition import RecognitionClient, Image, Label, Task


def get_client(request):
    token = request.config.getoption("--token")
    client = RecognitionClient(token)
    return client


def test_client_non_existing(request):
    """Tests an API call for non existing Token"""
    token = 'non-existing-token'
    client = RecognitionClient(token)
    response, status = client.get_all_tasks()

    assert response is None


def test_client_existing(request):
    """Tests an API call for existing Token"""
    client = get_client(request)
    tasks, status = client.get_all_tasks()

    assert isinstance(tasks, list)
    if len(tasks):
        assert isinstance(tasks[0], Task)
        assert hasattr(tasks[0], 'id')
        assert hasattr(tasks[0], 'name')
        assert hasattr(tasks[0], 'type')
        assert hasattr(tasks[0], 'production_version')


def test_client_all_labels(request):
    """Tests an API call for all labels"""
    client = get_client(request)
    labels, status = client.get_all_labels()

    assert isinstance(labels, list)
    if len(labels):
        assert isinstance(labels[0], Label)
        assert hasattr(labels[0], 'id')
        assert hasattr(labels[0], 'name')


def test_client_create_task_label_image(request):
    """Tests an API call for creating task, label, image."""
    client = get_client(request)
    labels_before, status = client.get_all_labels()
    task, status = client.create_task('Test-Task-In-Vize-X-1')
    task1, status = client.get_task(task.id)
    label, status = task.create_label('Test-Task-In-Vize-Label-X-1')
    labels1, status = task1.get_labels()
    images0, n_page, status = label.get_training_images()
    image, status = label.upload_image('ximilar.png')
    images1, n_page, status = label.get_training_images()
    label.detach_image(image.id)
    images0_new, n_page, status = label.get_training_images()

    task.detach_label(label.id)
    client.delete_task(task.id)
    client.delete_label(label.id)
    client.delete_image(image.id)
    labels_after, status = client.get_all_labels()

    no_task, no_task_status = client.get_task(task.id)
    no_image, no_image_status = client.get_label(label.id)
    no_label, no_label_status = client.get_label(label.id)

    assert(no_task, None)
    assert(no_image, None)
    assert(no_label, None)

    # check creating of Task, Label and Image
    assert isinstance(task, Task)
    assert isinstance(label, Label)
    assert isinstance(image, Image)

    # check the get endpoints
    assert task.id == task1.id
    assert label.id == labels1[0].id

    # check the creating of label and image
    assert len(labels_before) == len(labels_after)
    assert len(labels1) == 1
    assert len(images0) == 0
    assert len(images1) == 1
    assert len(images0_new) == 0


def test_classify(request):
    """Tests an API call for one of our label"""
    client = get_client(request)
    tasks, status = client.get_all_tasks()
    task = tasks[0]

    result = task.classify([{"_file": "ximilar.png"}])
    assert isinstance(result, dict)





from ximilar.client import RestClient
from ximilar.client.constants import RECORDS

FASHION_TAGGING_ENDPOINT = 'tagging/fashion/v2/tags'
GENERIC_TAGGING_ENDPOINT = 'tagging/generic/v2/tags'


class TaggingClient(RestClient):
    def tags(self, records, endpoint):
        records = self.preprocess_records(records)
        return self.post(endpoint, data={RECORDS: records})


class FashionTaggingClient(TaggingClient):
    def tags(self, records):
        return super().tags(records, FASHION_TAGGING_ENDPOINT)


class GenericTaggingClient(TaggingClient):
    def tags(self, records):
        return super().tags(records, GENERIC_TAGGING_ENDPOINT)



from ximilar.client import RestClient
from ximilar.client.constants import RECORDS

DOMINANT_COLOR_GENERIC_ENDPOINT = 'generic_photos/v2/dominantcolor'
DOMINANT_COLOR_PRODUCT_ENDPOINT = 'product_photos/v2/dominantcolor'


class DominantColorClient(RestClient):
    def dominantcolor(self, records, endpoint):
        records = self.preprocess_records(records)
        return self.post(endpoint, data={RECORDS: records})


class DominantColorGenericClient(DominantColorClient):
    def dominantcolor(self, records):
        return super().dominantcolor(records, DOMINANT_COLOR_GENERIC_ENDPOINT)


class DominantColorProductClient(DominantColorClient):
    def dominantcolor(self, records):
        return super().dominantcolor(records, DOMINANT_COLOR_PRODUCT_ENDPOINT)




# General
RECORDS = 'records'
FILE = '_file'
URL = '_url'
BASE64 = '_base64'
IMG_DATA = '_img_data'
RESULT_OK = {'status': 'OK'}

# Recognition/Vize related
RESULTS = 'results'
ID = 'id'
NAME = 'name'
TASK = 'task'
TASKS_COUNT = 'tasksCount'



from setuptools import setup, find_packages
import os

with open('requirements.txt') as f:
    install_requirements = f.read().splitlines()

if os.environ.get('CI_COMMIT_TAG'):
    version = os.environ['CI_COMMIT_TAG']
else:
    version = os.environ['CI_JOB_ID']


setup(name='ximilar-client',
    version=version,
    description='The Ximilar App and Vize.ai Client.',
    url='http://ximilar.com/',
    author='Michal Lukac, David Novak and Ximilar.com Team',
    author_email='tech@ximilar.com',
    license='Apache 2.0',
    packages=find_packages(),
    keywords='machine learning, multimedia, json, rest, data',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3.4',
        'Topic :: Scientific/Engineering :: Artificial Intelligence'
    ],
    install_requires=install_requirements,
    include_package_data=True,
    zip_safe=False,
    namespace_packages=["ximilar"])



from ximilar.client import RestClient
from ximilar.client.constants import TASK, NAME, ID, RESULTS, TASKS_COUNT, RESULT_OK

LABEL_ENDPOINT = 'recognition/v2/label/'
TASK_ENDPOINT = 'recognition/v2/task/'
IMAGE_ENDPOINT = 'recognition/v2/training-image/'
CLASSIFY_ENDPOINT = 'recognition/v2/classify/'


class RecognitionClient(RestClient):
    def get_task(self, task_id):
        """
        Getting task by id.
        :param task_id: id of task
        :return: Task object
        """
        task_json = self.get(TASK_ENDPOINT + task_id)
        if 'id' not in task_json:
            return None, {'status': 'Not Found'}
        return Task(self.token, self.endpoint, task_json), RESULT_OK

    def get_all_tasks(self, suffix=''):
        """
        Get all tasks of the user(user is specified by client key).
        :return: List of Tasks
        """
        url, tasks = TASK_ENDPOINT+suffix, []

        while True:
            result = self.get(url)

            if RESULTS not in result:
                return None, {'status': 'unexpected error'}

            for task_json in result[RESULTS]:
                tasks.append(Task(self.token, self.endpoint, task_json))

            if result['next'] is None:
                break

            url = result['next'].replace(self.endpoint, "").replace(self.endpoint.replace("https", "http"), "")

        return tasks, RESULT_OK

    def get_task_by_name(self, name):
        tasks, result = self.get_all_tasks()

        if result['status'] == 'OK':
            for task in tasks:
                if task.name == name:
                    return task, result
        else:
            return None, result
        return None, {'status': 'Task with this name not found!'}

    def delete_task(self, task_id):
        """
        Delete the task from the db.
        :param task_id: identification of task
        :return: None
        """
        return self.delete(TASK_ENDPOINT + task_id + '/')

    def create_task(self, name):
        """
        Create task with given name.
        :param name: name of the task
        :return: Task object
        """
        task_json = self.post(TASK_ENDPOINT, data={NAME: name})
        if 'id' not in task_json:
            msg = task_json['detail'] if 'detail' in task_json else 'unexpected error'
            return None, {'status': msg}
        return Task(self.token, self.endpoint, task_json), RESULT_OK

    def create_label(self, name):
        """
        Create label with given name.
        :param name: name of the label
        :return: Label object
        """
        label_json = self.post(LABEL_ENDPOINT, data={NAME: name})
        if 'id' not in label_json:
            return None, {'status': 'unexpected error'}
        return Label(self.token, self.endpoint, label_json), RESULT_OK

    def get_all_labels(self, suffix=''):
        """
        Get all labels of the user(user is specified by client key).
        :return: List of labels
        """
        url, labels = LABEL_ENDPOINT+suffix, []

        while True:
            result = self.get(url)

            for label_json in result[RESULTS]:
                labels.append(Label(self.token, self.endpoint, label_json))

            if result['next'] is None:
                break

            url = result['next'].replace(self.endpoint, "").replace(self.endpoint.replace("https", "http"), "")

        return labels, RESULT_OK

    def get_label(self, label_id):
        label_json = self.get(LABEL_ENDPOINT + label_id)
        if ID not in label_json:
            return None, {'status': 'label with id not found'}
        return Label(self.token, self.endpoint, label_json), RESULT_OK

    def get_image(self, image_id):
        image_json = self.get(IMAGE_ENDPOINT + image_id)
        if ID not in image_json:
            return None, {'status': 'image with id not found'}
        return Image(self.token, self.endpoint, image_json), RESULT_OK

    def delete_label(self, label_id):
        self.delete(LABEL_ENDPOINT + label_id)

    def delete_image(self, image_id):
        return self.delete(IMAGE_ENDPOINT + image_id)

    def upload_image(self, file_path=None, base64=None, label_ids=[]):
        files, data = None, None

        if file_path:
            files = {'img_path': open(file_path, 'rb')}
        if base64:
            data = {'base64': base64.decode("utf-8")}

        image_json = self.post(IMAGE_ENDPOINT, files=files, data=data)
        image, status = self.get_image(image_json['id'])

        for label_id in label_ids:
            image.add_label(label_id)
        return image, RESULT_OK


class Task(RecognitionClient):
    def __init__(self, token, endpoint, task_json):
        super(Task, self).__init__(token, endpoint)

        self.id = task_json[ID]
        self.name = task_json[NAME]
        self.type = task_json['type']
        self.production_version = task_json['production_version']

    def __str__(self):
        return self.id

    def train(self):
        """
        Create new training in ximilar.ai.
        :return: None
        """
        return self.post(TASK_ENDPOINT+self.id+'/train/')

    def delete_task(self):
        """
        Delete the task from ximilar.
        :return: None
        """
        super(Task, self).delete_task(self.id)

    def get_labels(self):
        """
        Get labels of this task.
        :return: list of Labels
        """
        if 'labels' in self.cache:
            return self.cache['labels'], RESULT_OK
        else:
            labels, result = self.get_all_labels(suffix='?task=' + self.id)

            if result['status'] == 'OK':
                self.cache['labels'] = labels

            return self.cache['labels'], result

    def get_label_by_name(self, name):
        """
        Get label of this task by name.
        """
        labels, result = self.get_labels()
        if result['status'] == 'OK':
            for label in labels:
                if label.name == name:
                    return label, RESULT_OK
        else:
            return None, result

        return None, {'status': 'Label with this name not found!'}

    def classify(self, records, version=None):
        """
        Takes the images and calls the ximilar client for classifying these images on the task.

        Usage:
            client = VizeRestClient('your-token')
            task = client.get_task('')
            result = task.classify({'_url':'http://example.com/client.jpg'})

        :param records: array of json/dicts [{'_url':'url-path'}, {'_file': ''}, {'_base64': 'base64encodeimg'}]
        :param version: optional(integer of specific version), default None/production_version
        :return: json response
        """
        records = self.preprocess_records(records)

        # version is default set to None, so ximilar will determine which one to take
        data = {'records': records, 'task_id': self.id, 'version': version}
        return self.post(CLASSIFY_ENDPOINT, data=data)

    def create_label(self, name):
        """
        Creates label and adds it to this task.
        :param name:
        :return:
        """
        label, result = super(Task, self).create_label(name)
        self.add_label(label.id)
        return label, RESULT_OK

    def add_label(self, label_id):
        """
        Add label to this task. If the task was already trained then it is frozen and you are not able to add new label.
        :param label_id: identification of label
        :return: json/dict result
        """
        return self.post(TASK_ENDPOINT + self.id +'/add-label/', data={'label_id': label_id})

    def detach_label(self, label_id):
        """
        Remove/Detach label from the task. If the task was already trained then it is frozen and you are not able to remove it.
        :param label_id: identification of label
        :return: json/dict result
        """
        return self.post(TASK_ENDPOINT + self.id +'/remove-label/', data={'label_id': label_id})


class Label(RecognitionClient):
    def __init__(self, token, endpoint, label_json):
        super(Label, self).__init__(token, endpoint)

        self.id = label_json[ID]
        self.name = label_json[NAME]
        self.tasks_count = label_json[TASKS_COUNT] if TASKS_COUNT in label_json else 0

    def __str__(self):
        return self.id

    def wipe_label(self):
        """
        Delete label and all images associated with this label.
        :return: None
        """
        self.delete(LABEL_ENDPOINT + self.id + '/wipe')

    def merge_label(self, label_b):
        """
        Get all photos of label_b and remove the label_b. Add this label to all the photos.
        This will lead to merging these two labels.
        """
        next_page, images = None, []
        while True:
            images_m, next_page, status = label_b.get_training_images(page_url=next_page)
            for image in images_m:
                 images.append(image)

            if not next_page:
                break

        for image in images:
            image.remove_label(label_b.id)
            image.add_label(self.id)

    def get_training_images(self, page_url=None):
        """
        Get paginated result of images for specific label.

        :param page_url: optional, select the specific page of images, default first page
        :return: (list of images, next_page)
        """
        url = page_url.replace(self.endpoint, "").replace(self.endpoint.replace("https", "http"), "") if page_url else IMAGE_ENDPOINT + '?label='+self.id
        result = self.get(url)
        return [Image(self.token, self.endpoint, image_json) for image_json in result[RESULTS]], result['next'], RESULT_OK

    def upload_image(self, file_path=None, base64=None):
        """
        Upload image to the ximilar.ai and adding label to this image.
        :param file_path: local path to the file
        :return: None
        """
        return super(Label, self).upload_image(file_path=file_path, base64=base64, label_ids=[self.id])

    def detach_image(self, image_id):
        """
        Remove/Detach label from the image.
        :param label_id: id of label
        :return: result
        """
        return self.post(IMAGE_ENDPOINT + image_id + '/remove-label/', data={'label_id': self.id})

    def delete_label(self):
        """
        Delete the image from ximilar.
        :return: None
        """
        super(Label, self).delete_label(self.id)


class Image(RecognitionClient):
    def __init__(self, token, endpoint, image_json):
        super(Image, self).__init__(token, endpoint)

        self.id = image_json[ID]
        self.thumb_img_path = image_json['thumb_img_path']

    def __str__(self):
        return self.thumb_img_path

    def delete_image(self):
        """
        Delete the image from ximilar.
        :return: None
        """
        super(Image, self).delete_image(self.id)

    def get_labels(self):
        """
        Get labels assigned to this image.
        :return: list of Labels
        """
        return [Label(self.token, self.endpoint, label) for label in self.get(IMAGE_ENDPOINT + self.id)['labels']], RESULT_OK

    def add_label(self, label_id):
        """
        Add label to the image.
        :param label_id: id of label
        :return: result
        """
        return self.post(IMAGE_ENDPOINT + self.id + '/add-label/', data={'label_id': label_id})

    def detach_label(self, label_id):
        """
        Remove/Detach label from the image.
        :param label_id: id of label
        :return: result
        """
        return self.post(IMAGE_ENDPOINT + self.id + '/remove-label/', data={'label_id': label_id})